==> fspathtree-0.7-py3-none-any.whl: fspathtree/fspathtree.py <==
from pathlib import PurePosixPath
import re
from inspect import signature
import collections
import copy

class PathGoesAboveRoot(Exception):
  pass
class CannotCreateParentNode(Exception):
  pass
class InvalidIndexError(Exception):
  pass
class UnrecognizedParentNodeType(Exception):
  def __init__(self,node_type,remaining_parts):
    self.node_type = node_type
    self.remaining_parts = remaining_parts

class NodeError(Exception):
  pass

class fspathtree:
  """A small class that wraps a tree data struction and allow accessing the nested elements using filesystem-style paths."""
  MappingNodeType = dict
  SequenceNodeType = list
  DefaultNodeType = MappingNodeType
  PathType = PurePosixPath

  def __init__(self,tree=None,root=None,abspath='/'):
    # we don't want to created "nested" fspathtree objects
    # beccause we want them to be light wrappers around regular
    # types. so, if tree is an fspathtree instance, we want to
    # wrap its tree instead
    if isinstance(tree,fspathtree):
        tree = tree.tree
    self.tree = tree if tree is not None else self.DefaultNodeType()
    self.root = root if root is not None else self.tree

    self.abspath = self.PathType(abspath)
    
    if self.tree == self.root and abspath != '/':
      raise RuntimeError("fspathtree: tree initialized with a root, but abspath is not '/'.")

    if self.tree != self.root and abspath == '/':
      raise RuntimeError("fspathtree: tree initialized with an abspath '/', but the tree and root are not the same.")

    self.get_all_leaf_node_paths = self._instance_get_all_leaf_node_paths
    self.find = self._instance_find

  @staticmethod
  def _make_node_type_for_key(key):
      if key.isnumeric():
          return fspathtree.SequenceNodeType()
      return fspathtree.MappingNodeType()

  @staticmethod
  def is_leaf(key,node):
    if type(node) in [str,bytes]:
      return True

    if isinstance(node,collections.abc.Mapping) or isinstance(node,collections.abc.Sequence):
      return False

    return True

  # Public Instance API

  def __getitem__(self,path,wrap_branch_nodes=True):
    path = self._make_path(path)

    if path.is_absolute():
      node = fspathtree.getitem(self.root,path)
    else:
      try:
        node = fspathtree.getitem(self.tree,path)
      except PathGoesAboveRoot as e:
        # if the key references a node above th root,
        # try again with the root tree.
        if self.abspath == self.PathType("/"):
          raise e
        node = fspathtree.getitem(self.root,(self.abspath/path))

    # if the item is an indexable node, we want to wrap it in an fspathtree before returning.
    if fspathtree.is_leaf(path,node) or wrap_branch_nodes is False:
      return node
    else:
      return fspathtree(node,root=self.root,abspath=(self.abspath/path).as_posix())


  def __setitem__(self,key,value):
    '''
    Set the element located at `key` to `value`.

    The `key` is treated as a filesystem-like path, with parts separated by '/' characters.
    If the path represented by `key` includes "parents" that do not exists, they will be created.
    '''
    path = self._make_path(key)

    if path.is_absolute():
      fspathtree.setitem(self.root,path,value)
      return

    # path is relative
    # first try to set the item from local tree
    # if a PathGoesAboveRoot exception is thrown, then
    # we can check to see if the path refers to an path in the
    # root tree
    try:
      fspathtree.setitem(self.tree,path,value)
    except PathGoesAboveRoot as e:
      if self.abspath == self.PathType("/"):
        raise e
      fspathtree.setitem(self.root,(self.abspath/path),value)

  def __contains__(self,key):
    try:
      self[key]
      return True
    except:
      return False

  def __len__(self):
    return len(self.tree)

  def update(self,*args,**kwargs):
    # if _any_ of the arguments are an fspathtree, then
    # we want to implemented a nested update.
    if any( [ isinstance(a,fspathtree) for a in args] ):
        for a in args:
            if type(a) is dict:
                a = fspathtree(a) # convert all non-trees to trees, so we can subscript them with path keys

            for key in a.get_all_leaf_node_paths():
                self[key] = a[key]

        for key in kwargs:
            self[key] = kwargs[key]


    else:
    # otherwise, we just use the 
        self.tree.update(*args,**kwargs)

  def path(self):
    return self.normalize_path(self.abspath)

  def get(self,path,default_value):
    '''
    Returns the value of the node references by path, or a default value if the node does not exist.
    '''
    try:
      return self[path]
    except KeyError:
      return default_value


  # this is used to allow the same name for instance and static methods
  def _instance_get_all_leaf_node_paths(self, transform = None, predicate=None):
    return fspathtree.get_all_leaf_node_paths(self.tree,transform,predicate)


  def _instance_find(self,pattern):
    return fspathtree.find(self.tree,pattern)



  # Public Static API



  @staticmethod
  def normalize_path(path,up="..",current="."):
    parts = fspathtree._normalize_path_parts( path.parts, up, current)
    if parts is None:
      return None
    return fspathtree.PathType(*parts)

  @staticmethod
  def getitem(tree,path):
    '''
    Given a tree and a path, returns the value of the node pointed to by the path. The path will be normalized first.

    path may be specified as a string, Path-like object, or list of path elements.
    '''
    original_path = copy.copy(path)
    path = fspathtree._make_path(path,normalize_path=False)
    # remove the '/' from the beginning of the path if it exists.
    if path.is_absolute():
      path = path.relative_to('/')
    if str(path) == '' or str(path) == '.':
      return tree

    try:
      return fspathtree._getitem_from_path_parts(tree,path.parts)
    except KeyError as e:
      msg = f"Could not find path element '{e.args[0]}' while parsing path '{original_path}'"
      raise KeyError(msg)
    except IndexError as e:
      msg = f"Could not find path element '{e.args[0]}' while parsing path '{original_path}'"
      raise IndexError(msg)
    except InvalidIndexError as e:
      msg = f"Error while parsing '{original_path}'." + str(e)
      raise IndexError(msg)
    except Exception as e:
      raise e



  @staticmethod
  def setitem(tree,path,value):
    '''
    Given a tree, a path, and a value, sets the value of the node pointed to by the path. If any level of the path does not
    exist, it is created. If the _root_ node type needs to be changed to set the value, then a new tree with the value set
    is returned.
    '''
    original_path = copy.copy(path)
    path = fspathtree._make_path(path,normalize_path=False)
    # remove the '/' from the beginning of the path if it exists.
    if path.is_absolute():
      path = path.relative_to('/')

    try:
      return fspathtree._setitem_from_path_parts(tree,path.parts,value)
    except KeyError as e:
      msg = f"Could not find path element '{e.args[0]}' while parsing path '{original_path}'"
      raise KeyError(msg)
    except IndexError as e:
      msg = f"Could not find path element '{e.args[0]}' while parsing path '{original_path}'"
      raise KeyError(msg)
    except InvalidIndexError as e:
      msg = f"Error while parsing path '{original_path}'. " + str(e)
      raise IndexError(msg)
    except Exception as e:
      raise e

  @staticmethod
  def get_all_leaf_node_paths(node,transform = None ,predicate = None):
    if transform is False:
      transform = None
    return fspathtree._get_all_leaf_node_paths(node,transform,predicate)

  @staticmethod
  def find(tree,pattern,as_string=False):
    return fspathtree.get_all_leaf_node_paths(tree,str if as_string else None,lambda p: p.match(pattern))


  # Private Methods

  @staticmethod
  def _make_path(key,normalize_path=False):
    '''
    Given a string, bytes array, integer, or list of path elements;  return a PathType object representing the path.
    '''
    if type(key) in (list,tuple):
      path = fspathtree.PathType(*key)
    else:
      if type(key) in (str,bytes):
        key = re.sub(r'^\/+','/',key) # replace multiple '/' at front with a single '/'. i.e. // -> /

      if type(key) in (int,):
        key = str(key)
      path = fspathtree.PathType(key)

    if normalize_path:
      path = fspathtree.normalize_path(path)
      if path is None:
        raise PathGoesAboveRoot("fspathtree: Key path contains a parent reference (..) that goes above the root of the tree")

    return path

  @staticmethod
  def _normalize_path_parts(parts,up="..",current="."):
    '''
    Normalize a list of path parts. i.e. remove ".." and "." elements
    while maintaining the same path.
    '''

    if up not in parts and current not in parts:
      return parts

    norm_parts = list()
    for p in parts:
      if p == current:
        continue
      elif p == up:
        if len(norm_parts) < 1:
          return None
        del norm_parts[-1]
      else:
        norm_parts.append(p)

    return norm_parts

  @staticmethod
  def _getitem_from_path_parts(tree,path_parts):
    '''
    @param tree           A nested Mapping/Sequence (i.e. dict/list) object that is being accessed.
    @param path_parts     A list of path elements identifying the location of the element being accessed.
    '''
    # normalize the path before traversing
    path_parts = fspathtree._normalize_path_parts(path_parts)

    if path_parts is None:
      raise PathGoesAboveRoot("fspathtree: Key path contains a parent reference (..) that goes above the root of the tree")

    try:
        return fspathtree.__getitem_from_path_parts_imp(tree,path_parts)
    except UnrecognizedParentNodeType as e:
        # if at any point during the traversal we run into a node that we cannot figure out
        # how to subscript. we will throw an UnrecognizedParentNodeType esception. However,
        # we don't want to throw that to the caller, becasue it will not have any information
        # about where in the path the error occured. So, we will catch this exception,
        # figure out where it happened, and throw a nicer error.
        error_node_path_parts = [ p for p in path_parts if p not in e.remaining_parts]
        raise NodeError(f"Unknown parent node type ({e.node_type}) found at '{'/'.join(error_node_path_parts)}' while traversing path '{'/'.join(path_parts)}'. This likely means you are trying to access an element beyond a current leaf node.")

  def __getitem_from_path_parts_imp(tree,path_parts):
    if isinstance(tree,collections.abc.Mapping):
      if path_parts[0] in tree:
        node = tree[path_parts[0]]
      else:
        raise KeyError(path_parts[0])
    elif isinstance(tree,collections.abc.Sequence):
      if len(tree) > int(path_parts[0]):
        node = tree[int(path_parts[0])]
      else:
        raise IndexError(path_parts[0])
    else:
      raise UnrecognizedParentNodeType(type(tree), path_parts)

    if len(path_parts) == 1:
      return node
    else:
      return fspathtree.__getitem_from_path_parts_imp(node,path_parts[1:])


  @staticmethod
  def _setitem_from_path_parts(tree,path_parts,value):
    '''
    Set the value of a tree branch located at the path given by `path_parts` to the given value.
    Any missing "parents" of the path are created.

    @param tree           A nested Mapping/Sequenhe (i.e. dict/list) object that is being accessed.
    @param path_parts     A list of path elements identifying the location of the element being accessed.
    @param _root          Implementation detail. Do not use.
    '''

    # If _root is true, it means this is the _first_ call and we need to normalize the path
    path_parts = fspathtree._normalize_path_parts(path_parts)

    if path_parts is None:
      raise PathGoesAboveRoot("fspathtree: Key path contains a parent reference (..) that goes above the root of the tree")
    
    fspathtree._setitem_from_path_parts_imp(tree,path_parts,value)




  @staticmethod
  def _setitem_from_path_parts_imp(tree,path_parts,value):
    key = path_parts[0]
    # if the tree is a dict, then add the path as a key
    if isinstance(tree,collections.abc.Mapping):
      if len(path_parts) == 1:
        tree[key] = value
      else:
        if key not in tree:
          tree[key] = fspathtree._make_node_type_for_key(path_parts[1])
        result = fspathtree._setitem_from_path_parts_imp(tree[key],path_parts[1:],value)
        if result is not None:
            tree[key] = result
            fspathtree._setitem_from_path_parts_imp(tree[key],path_parts[1:],value)


    # if the tree is a dict, then add the path as an elment of the list with given index
    elif isinstance(tree,collections.abc.Sequence):
      # check that the path element is an integer.
      # if it is not, then we want to convert the list to a dict
      # and tell the caller to try again...
      if not key.isnumeric():
          raise InvalidIndexError(f"Non-numeric index '{key}' used to index Sequence node.")

      key = int(key)

      # if the list does not have enough elements
      # append None until it does
      while len(tree) <= key:
        tree.append(None)
      if len(path_parts) == 1:
        tree[key] = value
      else:
        if tree[key] is None:
          tree[key] = fspathtree._make_node_type_for_key(path_parts[1])
        result = fspathtree._setitem_from_path_parts_imp(tree[key],path_parts[1:],value)
        if result is not None:
            tree[key] = result
            fspathtree._setitem_from_path_parts_imp(tree[key],path_parts[1:],value)
    else:
      # if we are here, then it means this node has _previously_ been set to a scalar.
      # i.e.
      # t['/level1/level2'] = 1
      # t['/level1/level2/level3'] = 1 # level2 is not a dict or list!
      # we want to support this, so we need to convert the node. if the path
      # element is an int (or convertable to an int), then we want to replace
      # the node with a list, otherwise, we will replace it with a dict.
      # we will return an empty version of the new node type so that the caller (above)
      # can replace the node and try again.
      if path_parts[0].isnumeric():
          return fspathtree.SequenceNodeType()
      return fspathtree.MappingNodeType()
      raise CannotCreateParentNode(f"fspathree: unrecognized node type '{type(tree)}' is not Mapping or Sequence. Do not know how to set item.")




  @staticmethod
  def _get_all_leaf_node_paths(node, transform = None, predicate = None, current_path=PathType("/")):
    '''
    Returns a list containing the paths to all leaf nodes in the tree.
    '''
    if not fspathtree.is_leaf(current_path,node):
      try:
        for i in range(len(node)):
          yield from fspathtree._get_all_leaf_node_paths( node[i], transform, predicate, current_path / str(i))
      except:
        for k in node:
          yield from fspathtree._get_all_leaf_node_paths( node[k], transform, predicate, current_path / k)
    else:
      return_path = True
      if predicate is not None:
        num_args = len(signature(predicate).parameters)
        if num_args  == 1:
          return_path = predicate(current_path)
        elif num_args == 2:
          return_path = predicate(current_path,node)
        else:
          raise RuntimeError(f"fspathtree: Predicate function not supported. Predicates may take 1 or 2 arguments. Provided function takes {num_args}.")

      if return_path:
        if transform is None:
          yield current_path
        elif type(transform) == type:
          yield transform(current_path)
        else:
          num_args = len(signature(transform).parameters)
          if num_args == 1:
            yield transform(current_path)
          elif num_args == 2:
            yield transform(current_path,node)
          else:
            raise RuntimeError(f"fspathtree: Transform function not supported. Transforms may take 1 or 2 arguments. Provided function takes {num_args}.")

==> fspathtree-0.8-py3-none-any.whl: fspathtree/fspathtree.py <==
from pathlib import PurePosixPath
import re
from inspect import signature
import collections
import copy

class PathGoesAboveRoot(Exception):
  pass
class CannotCreateParentNode(Exception):
  pass
class InvalidIndexError(Exception):
  pass
class UnrecognizedParentNodeType(Exception):
  def __init__(self,node_type,remaining_parts):
    self.node_type = node_type
    self.remaining_parts = remaining_parts

class NodeError(Exception):
  pass

class fspathtree:
  """A small class that wraps a tree data struction and allow accessing the nested elements using filesystem-style paths."""
  MappingNodeType = dict
  SequenceNodeType = list
  DefaultNodeType = MappingNodeType
  PathType = PurePosixPath

  def __init__(self,tree=None,root=None,abspath='/'):
    # we don't want to created "nested" fspathtree objects
    # beccause we want them to be light wrappers around regular
    # types. so, if tree is an fspathtree instance, we want to
    # wrap its tree instead
    if isinstance(tree,fspathtree):
        tree = tree.tree
    self.tree = tree if tree is not None else self.DefaultNodeType()
    self.root = root if root is not None else self.tree

    self.abspath = self.PathType(abspath)
    
    if self.tree == self.root and abspath != '/':
      raise RuntimeError("fspathtree: tree initialized with a root, but abspath is not '/'.")

    if self.tree != self.root and abspath == '/':
      raise RuntimeError("fspathtree: tree initialized with an abspath '/', but the tree and root are not the same.")

    self.get_all_leaf_node_paths = self._instance_get_all_leaf_node_paths
    self.find = self._instance_find

  @staticmethod
  def _make_node_type_for_key(key):
      if key.isnumeric():
          return fspathtree.SequenceNodeType()
      return fspathtree.MappingNodeType()

  @staticmethod
  def is_leaf(key,node):
    if type(node) in [str,bytes]:
      return True

    if isinstance(node,collections.abc.Mapping) or isinstance(node,collections.abc.Sequence):
      return False

    return True

  # Public Instance API

  def __getitem__(self,path,wrap_branch_nodes=True):
    path = self._make_path(path)

    if path.is_absolute():
      node = fspathtree.getitem(self.root,path)
    else:
      try:
        node = fspathtree.getitem(self.tree,path)
      except PathGoesAboveRoot as e:
        # if the key references a node above th root,
        # try again with the root tree.
        if self.abspath == self.PathType("/"):
          raise e
        node = fspathtree.getitem(self.root,(self.abspath/path))

    # if the item is an indexable node, we want to wrap it in an fspathtree before returning.
    if fspathtree.is_leaf(path,node) or wrap_branch_nodes is False:
      return node
    else:
      return fspathtree(node,root=self.root,abspath=(self.abspath/path).as_posix())


  def __setitem__(self,key,value):
    '''
    Set the element located at `key` to `value`.

    The `key` is treated as a filesystem-like path, with parts separated by '/' characters.
    If the path represented by `key` includes "parents" that do not exists, they will be created.
    '''
    path = self._make_path(key)

    if path.is_absolute():
      fspathtree.setitem(self.root,path,value)
      return

    # path is relative
    # first try to set the item from local tree
    # if a PathGoesAboveRoot exception is thrown, then
    # we can check to see if the path refers to an path in the
    # root tree
    try:
      fspathtree.setitem(self.tree,path,value)
    except PathGoesAboveRoot as e:
      if self.abspath == self.PathType("/"):
        raise e
      fspathtree.setitem(self.root,(self.abspath/path),value)

  def __contains__(self,key):
    try:
      self[key]
      return True
    except:
      return False

  def __len__(self):
    return len(self.tree)

  def update(self,*args,**kwargs):
    # if _any_ of the arguments are an fspathtree, then
    # we want to implemented a nested update.
    if any( [ isinstance(a,fspathtree) for a in args] ):
        for a in args:
            if type(a) is dict:
                a = fspathtree(a) # convert all non-trees to trees, so we can subscript them with path keys

            for key in a.get_all_leaf_node_paths():
                self[key] = a[key]

        for key in kwargs:
            self[key] = kwargs[key]


    else:
    # otherwise, we just use the 
        self.tree.update(*args,**kwargs)

  def path(self):
    return self.normalize_path(self.abspath)

  def get(self,path,default_value):
    '''
    Returns the value of the node references by path, or a default value if the node does not exist.
    '''
    try:
      return self[path]
    except KeyError:
      return default_value


  # this is used to allow the same name for instance and static methods
  def _instance_get_all_leaf_node_paths(self, transform = None, predicate=None):
    root_path = fspathtree.PathType("")
    if self.tree == self.root:
        root_path = fspathtree.PathType("/")

    return fspathtree.get_all_leaf_node_paths(self.tree,transform,predicate,root_path=root_path)


  def _instance_find(self,pattern):
    root_path = fspathtree.PathType("")
    if self.tree == self.root:
        root_path = fspathtree.PathType("/")

    return fspathtree.find(self.tree,pattern,root_path=root_path)



  # Public Static API



  @staticmethod
  def normalize_path(path,up="..",current="."):
    parts = fspathtree._normalize_path_parts( path.parts, up, current)
    if parts is None:
      return None
    return fspathtree.PathType(*parts)

  @staticmethod
  def getitem(tree,path):
    '''
    Given a tree and a path, returns the value of the node pointed to by the path. The path will be normalized first.

    path may be specified as a string, Path-like object, or list of path elements.
    '''
    original_path = copy.copy(path)
    path = fspathtree._make_path(path,normalize_path=False)
    # remove the '/' from the beginning of the path if it exists.
    if path.is_absolute():
      path = path.relative_to('/')
    if str(path) == '' or str(path) == '.':
      return tree

    try:
      return fspathtree._getitem_from_path_parts(tree,path.parts)
    except KeyError as e:
      msg = f"Could not find path element '{e.args[0]}' while parsing path '{original_path}'"
      raise KeyError(msg)
    except IndexError as e:
      msg = f"Could not find path element '{e.args[0]}' while parsing path '{original_path}'"
      raise IndexError(msg)
    except InvalidIndexError as e:
      msg = f"Error while parsing '{original_path}'." + str(e)
      raise IndexError(msg)
    except Exception as e:
      raise e



  @staticmethod
  def setitem(tree,path,value):
    '''
    Given a tree, a path, and a value, sets the value of the node pointed to by the path. If any level of the path does not
    exist, it is created. If the _root_ node type needs to be changed to set the value, then a new tree with the value set
    is returned.
    '''
    original_path = copy.copy(path)
    path = fspathtree._make_path(path,normalize_path=False)
    # remove the '/' from the beginning of the path if it exists.
    if path.is_absolute():
      path = path.relative_to('/')

    try:
      return fspathtree._setitem_from_path_parts(tree,path.parts,value)
    except KeyError as e:
      msg = f"Could not find path element '{e.args[0]}' while parsing path '{original_path}'"
      raise KeyError(msg)
    except IndexError as e:
      msg = f"Could not find path element '{e.args[0]}' while parsing path '{original_path}'"
      raise KeyError(msg)
    except InvalidIndexError as e:
      msg = f"Error while parsing path '{original_path}'. " + str(e)
      raise IndexError(msg)
    except Exception as e:
      raise e

  @staticmethod
  def get_all_leaf_node_paths(node,transform = None ,predicate = None, root_path=PathType()):
    if transform is False:
      transform = None
    return fspathtree._get_all_leaf_node_paths(node,transform,predicate,root_path)

  @staticmethod
  def find(tree,pattern,as_string=False,root_path=PathType()):
    return fspathtree.get_all_leaf_node_paths(tree,str if as_string else None,lambda p: p.match(pattern),root_path=root_path)


  # Private Methods

  @staticmethod
  def _make_path(key,normalize_path=False):
    '''
    Given a string, bytes array, integer, or list of path elements;  return a PathType object representing the path.
    '''
    if type(key) in (list,tuple):
      path = fspathtree.PathType(*key)
    else:
      if type(key) in (str,bytes):
        key = re.sub(r'^\/+','/',key) # replace multiple '/' at front with a single '/'. i.e. // -> /

      if type(key) in (int,):
        key = str(key)
      path = fspathtree.PathType(key)

    if normalize_path:
      path = fspathtree.normalize_path(path)
      if path is None:
        raise PathGoesAboveRoot("fspathtree: Key path contains a parent reference (..) that goes above the root of the tree")

    return path

  @staticmethod
  def _normalize_path_parts(parts,up="..",current="."):
    '''
    Normalize a list of path parts. i.e. remove ".." and "." elements
    while maintaining the same path.
    '''

    if up not in parts and current not in parts:
      return parts

    norm_parts = list()
    for p in parts:
      if p == current:
        continue
      elif p == up:
        if len(norm_parts) < 1:
          return None
        del norm_parts[-1]
      else:
        norm_parts.append(p)

    return norm_parts

  @staticmethod
  def _getitem_from_path_parts(tree,path_parts):
    '''
    @param tree           A nested Mapping/Sequence (i.e. dict/list) object that is being accessed.
    @param path_parts     A list of path elements identifying the location of the element being accessed.
    '''
    # normalize the path before traversing
    path_parts = fspathtree._normalize_path_parts(path_parts)

    if path_parts is None:
      raise PathGoesAboveRoot("fspathtree: Key path contains a parent reference (..) that goes above the root of the tree")

    try:
        return fspathtree.__getitem_from_path_parts_imp(tree,path_parts)
    except UnrecognizedParentNodeType as e:
        # if at any point during the traversal we run into a node that we cannot figure out
        # how to subscript. we will throw an UnrecognizedParentNodeType esception. However,
        # we don't want to throw that to the caller, becasue it will not have any information
        # about where in the path the error occured. So, we will catch this exception,
        # figure out where it happened, and throw a nicer error.
        error_node_path_parts = [ p for p in path_parts if p not in e.remaining_parts]
        raise NodeError(f"Unknown parent node type ({e.node_type}) found at '{'/'.join(error_node_path_parts)}' while traversing path '{'/'.join(path_parts)}'. This likely means you are trying to access an element beyond a current leaf node.")

  def __getitem_from_path_parts_imp(tree,path_parts):
    if isinstance(tree,collections.abc.Mapping):
      if path_parts[0] in tree:
        node = tree[path_parts[0]]
      else:
        raise KeyError(path_parts[0])
    elif isinstance(tree,collections.abc.Sequence):
      if len(tree) > int(path_parts[0]):
        node = tree[int(path_parts[0])]
      else:
        raise IndexError(path_parts[0])
    else:
      raise UnrecognizedParentNodeType(type(tree), path_parts)

    if len(path_parts) == 1:
      return node
    else:
      return fspathtree.__getitem_from_path_parts_imp(node,path_parts[1:])


  @staticmethod
  def _setitem_from_path_parts(tree,path_parts,value):
    '''
    Set the value of a tree branch located at the path given by `path_parts` to the given value.
    Any missing "parents" of the path are created.

    @param tree           A nested Mapping/Sequenhe (i.e. dict/list) object that is being accessed.
    @param path_parts     A list of path elements identifying the location of the element being accessed.
    @param _root          Implementation detail. Do not use.
    '''

    # If _root is true, it means this is the _first_ call and we need to normalize the path
    path_parts = fspathtree._normalize_path_parts(path_parts)

    if path_parts is None:
      raise PathGoesAboveRoot("fspathtree: Key path contains a parent reference (..) that goes above the root of the tree")
    
    fspathtree._setitem_from_path_parts_imp(tree,path_parts,value)




  @staticmethod
  def _setitem_from_path_parts_imp(tree,path_parts,value):
    key = path_parts[0]
    # if the tree is a dict, then add the path as a key
    if isinstance(tree,collections.abc.Mapping):
      if len(path_parts) == 1:
        tree[key] = value
      else:
        if key not in tree:
          tree[key] = fspathtree._make_node_type_for_key(path_parts[1])
        result = fspathtree._setitem_from_path_parts_imp(tree[key],path_parts[1:],value)
        if result is not None:
            tree[key] = result
            fspathtree._setitem_from_path_parts_imp(tree[key],path_parts[1:],value)


    # if the tree is a dict, then add the path as an elment of the list with given index
    elif isinstance(tree,collections.abc.Sequence):
      # check that the path element is an integer.
      # if it is not, then we want to convert the list to a dict
      # and tell the caller to try again...
      if not key.isnumeric():
          raise InvalidIndexError(f"Non-numeric index '{key}' used to index Sequence node.")

      key = int(key)

      # if the list does not have enough elements
      # append None until it does
      while len(tree) <= key:
        tree.append(None)
      if len(path_parts) == 1:
        tree[key] = value
      else:
        if tree[key] is None:
          tree[key] = fspathtree._make_node_type_for_key(path_parts[1])
        result = fspathtree._setitem_from_path_parts_imp(tree[key],path_parts[1:],value)
        if result is not None:
            tree[key] = result
            fspathtree._setitem_from_path_parts_imp(tree[key],path_parts[1:],value)
    else:
      # if we are here, then it means this node has _previously_ been set to a scalar.
      # i.e.
      # t['/level1/level2'] = 1
      # t['/level1/level2/level3'] = 1 # level2 is not a dict or list!
      # we want to support this, so we need to convert the node. if the path
      # element is an int (or convertable to an int), then we want to replace
      # the node with a list, otherwise, we will replace it with a dict.
      # we will return an empty version of the new node type so that the caller (above)
      # can replace the node and try again.
      if path_parts[0].isnumeric():
          return fspathtree.SequenceNodeType()
      return fspathtree.MappingNodeType()
      raise CannotCreateParentNode(f"fspathree: unrecognized node type '{type(tree)}' is not Mapping or Sequence. Do not know how to set item.")




  @staticmethod
  def _get_all_leaf_node_paths(node, transform = None, predicate = None, root_path=PathType()):
    '''
    Returns a list containing the paths to all leaf nodes in the tree.
    '''
    if not fspathtree.is_leaf(root_path,node):
      try:
        for i in range(len(node)):
          yield from fspathtree._get_all_leaf_node_paths( node[i], transform, predicate, root_path / str(i))
      except:
        for k in node:
          yield from fspathtree._get_all_leaf_node_paths( node[k], transform, predicate, root_path / k)
    else:
      return_path = True
      if predicate is not None:
        num_args = len(signature(predicate).parameters)
        if num_args  == 1:
          return_path = predicate(root_path)
        elif num_args == 2:
          return_path = predicate(root_path,node)
        else:
          raise RuntimeError(f"fspathtree: Predicate function not supported. Predicates may take 1 or 2 arguments. Provided function takes {num_args}.")

      if return_path:
        if transform is None:
          yield root_path
        elif type(transform) == type:
          yield transform(root_path)
        else:
          num_args = len(signature(transform).parameters)
          if num_args == 1:
            yield transform(root_path)
          elif num_args == 2:
            yield transform(root_path,node)
          else:
            raise RuntimeError(f"fspathtree: Transform function not supported. Transforms may take 1 or 2 arguments. Provided function takes {num_args}.")
